import torch
import numpy as np
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
from IPython import display
display.set_matplotlib_formats('svg')
import torch.nn.functional as F

class BasicBlock(nn.Module):

    def __init__(self, in_c, out_c, stride = 1):
        super().__init__()
        
        self.downsample = False
        
        self.conv1 = nn.Conv2d(
                        in_channels=in_c, 
                        out_channels=out_c, 
                        kernel_size=(3,3), 
                        padding=(1,1), 
                        stride=stride,
                        bias=False)
                        
        self.conv2 = nn.Conv2d(
                        in_channels=out_c, 
                        out_channels=out_c, 
                        kernel_size=(3,3), 
                        padding=(1,1), 
                        stride=1,
                        bias=False)   

        if in_c != out_c or stride > 1:
            self.downsample = True
        
        if self.downsample:
            self.down = nn.Conv2d(
                in_c,
                out_c,
                kernel_size=1,
                padding=0,
                stride=stride
            )
                        
    def forward(self, x):        
        x_orig = x
        #print("x_orig.shape", x_orig.shape)
        x = F.relu(self.conv1(x))
        #print(x.shape)
        x = self.conv2(x)
        #print(x.shape)
        #dimensions must be adjusted, see model_graph
        if self.downsample:
            x_orig = self.down(x_orig)        
        #print("x_orig.shape", x_orig.shape)
        x = x + x_orig
        x = F.relu(x)
        return x

def get_my_resnet():

    model = nn.Sequential(
        nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False),
        nn.MaxPool2d(kernel_size = 3, stride = 2, padding = 1),
        #2_1
        BasicBlock(64, 64, 1),
        #2_2
        BasicBlock(64, 64, 1),
        #3_1 - down
        BasicBlock(64,128, 2)
    )
    return model

model = get_my_resnet()#BasicBlock(32, 64, 2)
print(model)

from torchsummary import summary

model = model.to("cpu")
test_tensor = torch.randn(1, 3, 224, 224)
print(model(test_tensor).shape)
summary(model, (3, 224, 224))

import torchvision

torch_resnet18 = torchvision.models.resnet18(weights=None)
summary(torch_resnet18, (3, 224, 224))
print(torch_resnet18(test_tensor).shape)

from torchview import draw_graph
architecture = 'MyResNet'
model_graph = draw_graph(model, input_size=(1,3,224,224), graph_dir ='TB' , expand_nested=True, graph_name=f'{architecture}',save_graph=True,filename=f'{architecture}')
model_graph.visual_graph

architecture = 'torch_resnet18'
model_graph = draw_graph(torch_resnet18, input_size=(1,3,224,224), graph_dir ='TB' , expand_nested=True, graph_name=f'{architecture}',save_graph=True,filename=f'{architecture}')
model_graph.visual_graph
